"""
django-bitfield
~~~~~~~~~~~~~~~
"""
from __future__ import absolute_import

from bitfield.models import Bit, BitHandler, CompositeBitField, BitField  # NOQA

default_app_config = 'bitfield.apps.BitFieldAppConfig'

VERSION = "2.2.0"
